/*
Copyright (c) 2016 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdint.h>
#include <stdio.h>
#include <syslog.h>
#include <X11/Xlib.h>
#include <X11/extensions/XShm.h>
#include "x11iface.h"
#include "config.h"
#include "bfont.h"
#include "txtren.h"
#include "menu.h"
#include "graymap.h"
#include "desk.h"
#include "mklist.h"
#include "window.h"
#include "retarded.h"
#include "main.h"
#include "mprintf.h"

static cfg_desk_t cfg;

static menu_t *winmenu, *itemmenu;
static char *selitem;
static char *desk;

static void restart_all();


static void debug_func(char *arg)
{
  //int K;
  //K= atoi(arg);
  printf("debug: %s\n",arg);
}

void execute_program(char *cmd, char *fn)
{
  char **argv;
  int status;
  pid_t cld;

  while(waitpid(-1, &status, WNOHANG)>0);

  if (!cmd) return ;
  if (cmd[0]==':')
  {
    char *pfx;
    pfx= ":debug";
    if (!strncmp(cmd,pfx,strlen(pfx)))
    {
       debug_func(cmd+strlen(pfx));
       return ;
    }
    return ;
  }
  if ((cld=fork()))
  {
     int status;
     if (cld>0) waitpid(cld,&status,0);
     return ;
  }
  // now we're in the child
  if (fork()) exit(0);
  
  cmd= strdup(cmd);
  // we do this split here because cmd may be modified
  argv= stupid_gnome_unquote_split(cmd,fn);
  if (argv[0] && argv[0][0])
     execvp(argv[0], argv);
  exit(0);
}

static char *cfghome()
{
  char *xdg_config_home, *dir;
  xdg_config_home= getenv("XDG_CONFIG_HOME");
  if (!xdg_config_home)
  {
    char *home;
    home= getenv("HOME");
    if (!home) return NULL;
    xdg_config_home= mprintf("%s/.config",home);
  } 
  else 
    xdg_config_home= strdup(xdg_config_home);

  dir= mprintf("%s/mudesk",xdg_config_home);
  mkdir(xdg_config_home, 0700);
  mkdir(dir, 0700);
  free(xdg_config_home);
  return dir;
}

static void read_config()
{
  char *dir, *fn;
  if (!(dir=cfghome())) return ;
  memset(&cfg,0,sizeof(cfg));
  fn= mprintf("%s/config",dir);
  cfg_parse_desk(fn,&cfg);
  free(fn);
  free(dir);
}

static void free_all_icons()
{
  int i;
  for(i=0;i<n_icons;i++) deskicon_free(icons[i]);
  if (icons) free(icons);
  n_icons= 0;
  icons= NULL;
}

static void read_desk()
{
  char *dir,*fn;
  if (!(dir=cfghome())) return ;
  free_all_icons();
  fn= mprintf("%s/desk",dir);
  read_desk_file(fn);
  free(fn);
}

void write_desk()
{
  char *dir,*fn;
  if (!(dir=cfghome())) return ;
  fn= mprintf("%s/desk",dir);
  write_desk_file(fn);
  free(fn);
}

static int x_error_handler(Display *disp, XErrorEvent *event)
{
static const struct { const char *name; int code; }
 xerrordef[]=
{
{"Unknown request",-1},
{"X_CreateWindow",1},
{"X_ChangeWindowAttributes",2},
{"X_GetWindowAttributes",3},
{"X_DestroyWindow",4},
{"X_DestroySubwindows",5},
{"X_ChangeSaveSet",6},
{"X_ReparentWindow",7},
{"X_MapWindow",8},
{"X_MapSubwindows",9},
{"X_UnmapWindow",10},
{"X_UnmapSubwindows",11},
{"X_ConfigureWindow",12},
{"X_CirculateWindow",13},
{"X_GetGeometry",14},
{"X_QueryTree",15},
{"X_InternAtom",16},
{"X_GetAtomName",17},
{"X_ChangeProperty",18},
{"X_DeleteProperty",19},
{"X_GetProperty",20},
{"X_ListProperties",21},
{"X_SetSelectionOwner",22},
{"X_GetSelectionOwner",23},
{"X_ConvertSelection",24},
{"X_SendEvent",25},
{"X_GrabPointer",26},
{"X_UngrabPointer",27},
{"X_GrabButton",28},
{"X_UngrabButton",29},
{"X_ChangeActivePointerGrab",30},
{"X_GrabKeyboard",31},
{"X_UngrabKeyboard",32},
{"X_GrabKey",33},
{"X_UngrabKey",34},
{"X_AllowEvents",35},
{"X_GrabServer",36},
{"X_UngrabServer",37},
{"X_QueryPointer",38},
{"X_GetMotionEvents",39},
{"X_TranslateCoords",40},
{"X_WarpPointer",41},
{"X_SetInputFocus",42},
{"X_GetInputFocus",43},
{"X_QueryKeymap",44},
{"X_OpenFont",45},
{"X_CloseFont",46},
{"X_QueryFont",47},
{"X_QueryTextExtents",48},
{"X_ListFonts",49},
{"X_ListFontsWithInfo",50},
{"X_SetFontPath",51},
{"X_GetFontPath",52},
{"X_CreatePixmap",53},
{"X_FreePixmap",54},
{"X_CreateGC",55},
{"X_ChangeGC",56},
{"X_CopyGC",57},
{"X_SetDashes",58},
{"X_SetClipRectangles",59},
{"X_FreeGC",60},
{"X_ClearArea",61},
{"X_CopyArea",62},
{"X_CopyPlane",63},
{"X_PolyPoint",64},
{"X_PolyLine",65},
{"X_PolySegment",66},
{"X_PolyRectangle",67},
{"X_PolyArc",68},
{"X_FillPoly",69},
{"X_PolyFillRectangle",70},
{"X_PolyFillArc",71},
{"X_PutImage",72},
{"X_GetImage",73},
{"X_PolyText8",74},
{"X_PolyText16",75},
{"X_ImageText8",76},
{"X_ImageText16",77},
{"X_CreateColormap",78},
{"X_FreeColormap",79},
{"X_CopyColormapAndFree",80},
{"X_InstallColormap",81},
{"X_UninstallColormap",82},
{"X_ListInstalledColormaps",83},
{"X_AllocColor",84},
{"X_AllocNamedColor",85},
{"X_AllocColorCells",86},
{"X_AllocColorPlanes",87},
{"X_FreeColors",88},
{"X_StoreColors",89},
{"X_StoreNamedColor",90},
{"X_QueryColors",91},
{"X_LookupColor",92},
{"X_CreateCursor",93},
{"X_CreateGlyphCursor",94},
{"X_FreeCursor",95},
{"X_RecolorCursor",96},
{"X_QueryBestSize",97},
{"X_QueryExtension",98},
{"X_ListExtensions",99},
{"X_ChangeKeyboardMapping",100},
{"X_GetKeyboardMapping",101},
{"X_ChangeKeyboardControl",102},
{"X_GetKeyboardControl",103},
{"X_Bell",104},
{"X_ChangePointerControl",105},
{"X_GetPointerControl",106},
{"X_SetScreenSaver",107},
{"X_GetScreenSaver",108},
{"X_ChangeHosts",109},
{"X_ListHosts",110},
{"X_SetAccessControl",111},
{"X_SetCloseDownMode",112},
{"X_KillClient",113},
{"X_RotateProperties",114},
{"X_ForceScreenSaver",115},
{"X_SetPointerMapping",116},
{"X_GetPointerMapping",117},
{"X_SetModifierMapping",118},
{"X_GetModifierMapping",119},
{"X_NoOperation",127},
{NULL, 0}};
   char g[2000];
   int i;
   for(i=1;xerrordef[i].name;i++)
      if (xerrordef[i].code==event->request_code) break;
   if (!xerrordef[i].name) i= 0;
   memset(g,0,sizeof(g));
   XGetErrorText(disp, event->error_code,g, sizeof(g)-1);
   syslog(LOG_NOTICE,"X11 error: [%d] %s. Resource: %x. Request: %s[%d.%d]",
                       event->error_code, g, 
                       (unsigned int) event->resourceid, 
                       xerrordef[i].name,
                       event->request_code, event->minor_code);
   return 1;
}


static void shift_args(int idx,int *argc,char **argv)
{
  int i;
  for(i=idx;i<*argc-1;i++)
   argv[i]= argv[i+1];
  *argc -= 1;
}

static int intarg(int *argc,char **argv, const char *sw,int *R)
{
  if (*argc==1) return 0;
  if (!strcmp(argv[1], sw))
  {
    shift_args(1, argc,argv);
    if (*argc==1) return 0;
    *R= atoi(argv[1]);
    shift_args(1, argc,argv);
    return 1;
  }
  return 0;
}

static int strarg(int *argc,char **argv, const char *sw,char** R)
{
  if (*argc==1) return 0;
  if (!strcmp(argv[1], sw))
  {
    shift_args(1, argc,argv);
    if (*argc==1) return 0;
    *R= argv[1];
    shift_args(1, argc,argv);
    return 1;
  }
  return 0;
}

static int swiarg(int *argc,char **argv, const char *sw,int *R)
{
  if (*argc==1) return 0;
  if (!strcmp(argv[1], sw))
  {
    shift_args(1, argc,argv);
    *R= 1;
    return 1;
  }
  return 0;
}

static cfg_desk_t cfg;
static int main_w, main_h;

static void refresh_desktop(int force)
{
  char buf[2000];
  deskfile_t **fils; int n_fils;
  fils= NULL;

  n_fils= 0;
  if (desk)
    scan_desktop_dir(desk, &n_fils, &fils);
  else
  {
    char *home;
    home= getenv("HOME"); if (!home) home= "/tmp";
    snprintf(buf,sizeof(buf), "%s/Desktop", home);
    buf[sizeof(buf)-1]= 0;
    scan_desktop_dir(buf, &n_fils, &fils);
  }

  if (refresh_icon_list(n_fils, fils, &cfg, main_w, main_h,force))
   write_desk();
  window_restart(&cfg);
}

static void window_menu(int idx, void *arg)
{
  switch(idx)
  {
    case 1: refresh_desktop(0); return ;
    case 2: restart_all(); return ;
  }
}

static void item_menu(int idx, void *arg)
{
  switch(idx)
  {
  case 1: refresh_desktop(0); return ;
  case 2: restart_all(); return ;
  case 3:
   { deskicon_t *ico;
     ico= make_icon_from_fn(selitem,&cfg);
     replace_icon(selitem, ico);
     free(selitem);
     selitem= NULL; return; }
  }
}

void show_window_menu(Window window, int x,int y)
{
  menu_display(winmenu, window, x, y);
}

void show_item_menu(Window window,int x,int y,char *fn)
{
  menu_display(itemmenu, window, x, y);
  selitem= strdup(fn);
}

static void make_menus()
{
  if (winmenu) menu_free(winmenu);
  if (itemmenu) menu_free(itemmenu);

  winmenu= menu_new(window_menu, NULL, "New Shortcut",
                    "Refresh All", "Restart", NULL);
  itemmenu= menu_new(item_menu, NULL,
                      "Edit", "Refresh All",
                      "Restart", NULL);
}


int main(int argc,char **argv)
{
  int test= 0;

  if (init_display(NULL)) return 1;
  openlog("mudesk", 0, LOG_USER); 
  signal(SIGCHLD,SIG_IGN);
  XSetErrorHandler(x_error_handler);
  read_config();
  font_init();
  init_builtin_font();
  graymap_init(x_red_shift,x_green_shift,x_blue_shift);
  menu_init(&cfg.menu_style);

  main_w= x_root_width;
  main_h= x_root_height;

  while(argc>1)
  {
    if (intarg(&argc,argv, "-w", &main_w)) continue;
    if (intarg(&argc,argv, "-h", &main_h)) continue;
    if (strarg(&argc,argv, "-d", &desk)) continue;
    if (swiarg(&argc,argv, "-t", &test)) continue;
    fprintf(stderr,"%s: unknown argument %s: skipping it\n",argv[0], argv[1]);
    shift_args(1,&argc,argv);
  }
  while(main_w<=0) main_w+= x_root_width;
  while(main_h<=0) main_h+= x_root_height;

  read_desk();

  init_window(&cfg,main_w,main_h,test);
  refresh_desktop(0);
  make_menus();

  while(1)
  {
    XEvent event;
    XNextEvent(x_display, &event);
    handle_registered_event(&event);
  }

  closelog();
  return 0;
}

static void restart_all()
{
  read_config();
  menu_init(&cfg.menu_style);
  refresh_desktop(1);
  make_menus();
}
