/*
Copyright (c) 2016 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#line 418 "/home/dodo/tmp/mudesk-1.0/desk/mklist.u"
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include "ptrbuf.h"
#include "txtren.h"
#include "config.h"
#include "desk.h"
#include "mklist.h"
#include "image.h"
#include "rect.h"
#include "mprintf.h"

static u_font_t *desk_font;
#line 132 "/home/dodo/tmp/mudesk-1.0/desk/mklist.u"

#define arrsiz(X) (sizeof(X)/sizeof(X[0]))
#define n_builtin_ftype arrsiz(builtin_ftype)

#line 22 "mklist.c"

static deskicon_t* make_icon(deskfile_t *fil, cfg_desk_t *cfg);
static deskicon_t *create_icon
  (cfg_desk_t *cfg, char *ico1,char *ico2, char *cmd, char *fn, uint32_t mtime);
static void render_icon(uint32_t *D, int W,int H, int iW, int iH, uint32_t *iD);
static void render_text
  (int fg, uint32_t *D,int W,int H,int tW,int tH, uint8_t *tD);
static char *extract_name(char *str);
static deskicon_t *make_icon_from_dir(deskfile_t *fil,cfg_desk_t *cfg);
static deskicon_t* make_icon_from_shortcut
   (deskfile_t *fil, cfg_desk_t *cfg);
static int isspace(int chr);
static char *trim_spaces(char *str);
static char *find_icon_by_name(cfg_desk_t *cfg, char *name);
static int has_suffix(char *str,char *sfx);
static int int_setting(int given, int setting, int defval, int min, int max);
static void find_file_type(char *fn, int nft, cfg_file_type_t *ft, 
                           char **Ricon, char **Rcmd);
#line 98 "/home/dodo/tmp/mudesk-1.0/desk/mklist.u"

static char *builtin_img_ext[]= { ".jpg", ".jpeg", ".png", ".gif" };
static char *builtin_pdf_ext[]= { ".pdf" };
static char *builtin_c_ext[]= { ".c", ".cc", ".cpp", ".cxx",
                                ".h", ".hh", ".hxx" };
static char *builtin_html_ext[]= { ".html", ".htm" };

static cfg_file_type_t builtin_ftype[]=
{
  { ":file-image.png", 1,
    "xdg-open %f", 1,
    builtin_img_ext,
    arrsiz(builtin_img_ext), arrsiz(builtin_img_ext)
  },
  {
    ":file-pdf.png", 1,
    "xdg-open %f", 1,
    builtin_pdf_ext, 1, 1 
  },
  {
    ":file-c.png", 1,
    "xdg-open %f", 1,
    builtin_c_ext,
    arrsiz(builtin_c_ext), arrsiz(builtin_c_ext)
  },
  {
    ":file-html.png", 1,
    "xdg-open %f", 1,
    builtin_html_ext,
    arrsiz(builtin_html_ext), arrsiz(builtin_html_ext)
  }
};

#line 8 "/home/dodo/tmp/mudesk-1.0/desk/mklist.u"

int refresh_icon_list(int n_fil, deskfile_t **fil,
                       cfg_desk_t *cfg,
                       int spc_w,int spc_h, int force)
{
  ptrbuf_t *icolist;
  ptrbuf_t *filist;
  int i,j;
  size_t nn;
  deskicon_t *IC;
  int change;

  change= 0;
  icolist= ptrbuf_new(50);
  filist= ptrbuf_new(50);

  spc_init(spc_w,spc_h);

  for(i=0;i<n_fil;i++)
  {
    int found;
    found= 0;
    for(j=0;j<n_icons;j++)
     if (!strcmp(icons[j]->fn,fil[i]->name))
     {
       if (!force && icons[j]->mtime==fil[i]->mtime)
       {
         found= 1;
         ptrbuf_put(icolist, icons[j]);
         spc_remove(icons[j]->x,icons[j]->y,icons[j]->w,icons[j]->h);
       }
       else
       {
         fil[i]->aux1= icons[j]->x;
         fil[i]->aux2= icons[j]->y;
         deskicon_free(icons[j]);
         change= 1;
       }
       icons[j]= icons[n_icons-1];
       n_icons--;
       break;
     }
    if (found)
       deskfile_free(fil[i]);
    else
       ptrbuf_put(filist, fil[i]);

    fil[i]= fil[n_fil-1];
    n_fil--;
    i--;
  }
  if (fil) free(fil);
  fil= NULL;

  ptrbuf_free(filist, (void***) &fil, &nn);
  n_fil= nn;
  for(i=0;i<n_fil;i++)
  {
    IC= make_icon(fil[i],cfg);
    if (IC) 
    {
      if (fil[i]->aux1 && fil[i]->aux2)
      {
        IC->x= fil[i]->aux1;
        IC->y= fil[i]->aux2;
        spc_remove(IC->x,IC->y,IC->w,IC->h);
      }
      else
      {
        // place the icon, it's a brand new one.
        spc_alloc(&IC->x, &IC->y, IC->w, IC->h);
      }
      change= 1;
      ptrbuf_put(icolist, IC);
    }
    deskfile_free(fil[i]);
  }
  if (fil) free(fil);

   // these are the extra icons left in the icon array.
   // there is no corresponding file in the desktop directory.
  for(j=0;j<n_icons;j++) deskicon_free(icons[j]);
  if (icons) free(icons);

  ptrbuf_free(icolist, (void***) &icons, &nn);
  n_icons= nn;
  return change;
}

#line 447 "/home/dodo/tmp/mudesk-1.0/desk/mklist.u"

deskicon_t *make_icon_from_fn(char *fn,cfg_desk_t *cfg)
{
  deskfile_t *fil;
  deskicon_t *ico;
  if (!(fil=stat_file(fn))) return NULL;
  ico= make_icon(fil, cfg);
  deskfile_free(fil);
  return ico;
}

#line 137 "/home/dodo/tmp/mudesk-1.0/desk/mklist.u"

static deskicon_t* make_icon(deskfile_t *fil, cfg_desk_t *cfg)
{
  char *icon1, *cmd;
  char *icon2;

  if (fil->type=='D') return make_icon_from_dir(fil,cfg);
  if (has_suffix(fil->name,".desktop"))
  {
    deskicon_t *ico;
    ico= make_icon_from_shortcut(fil,cfg);
    if (ico) return ico;
  }

  icon1= NULL; cmd= NULL;
  if (cfg->N_file_type)
    find_file_type(fil->name, cfg->N_file_type, cfg->file_type,&icon1, &cmd);
  find_file_type(fil->name, n_builtin_ftype, builtin_ftype, &icon2, 
                                                           cmd ? NULL :&cmd);
  return create_icon(cfg, icon1, icon2, cmd, fil->name, fil->mtime);
}

static deskicon_t *create_icon
  (cfg_desk_t *cfg, char *ico1,char *ico2, char *cmd, char *fn, uint32_t mtime)
{
  uint32_t icow, icoh, *icod;
  uint32_t tw, th;
  uint8_t  *td;
  char *name;
  int icosiz, maxw, pad_mid;
  deskicon_t *ii;
  int w,h;
  uint32_t *data;

  icosiz= int_setting(cfg->G_icon_size, cfg->icon_size, 32, 0, 256);
  maxw=   int_setting(cfg->G_max_width, cfg->max_width, 2*icosiz, 0, 256);
  pad_mid=int_setting(cfg->G_pad_mid,   cfg->pad_mid, 6, 0, 32);

  if (load_icon(ico1, &icow, &icoh, &icod, icosiz))
     load_icon(ico2, &icow, &icoh, &icod, icosiz);
  name= extract_name(fn); 
  if (!desk_font)
  {
    int fosi; char *fona;
    if (cfg->G_font_name) fona= cfg->font_name;
                     else fona= ":default.ttf";
    fosi= int_setting(cfg->G_font_size, cfg->font_size, 14, 0, 32);
    desk_font= font_new(fona, fosi);
    if (!desk_font) desk_font= font_new(":default.ttf", 14);
  }
  text_render(desk_font, name, &tw, &td);
  th= desk_font->height;

  w= tw > icow ? tw : icow;
  w= w > maxw ? maxw : w;
  h= icoh + pad_mid + th;

  data= calloc(sizeof(uint32_t),w*h);

  render_icon(data, w, h, icow, icoh, icod);
  render_text(cfg->G_foreground ? cfg->G_foreground: 0xFFffff, 
              data+w*(h-th), w, th, tw, th, td);

  ii= calloc(1,sizeof(*ii));
  ii->w= w; 
  ii->h= h;
  ii->ico= data;
  ii->fn= strdup(fn);
  ii->mtime= mtime;

  if (!cmd) cmd= "xdg-open %f";
  ii->cmd= strdup(cmd);

  free(td);
  free(icod);
  return ii;
}

#line 216 "/home/dodo/tmp/mudesk-1.0/desk/mklist.u"

static void render_icon(uint32_t *D, int W,int H, int iW, int iH, uint32_t *iD)
{
  int sx, sy, dx, dy, cw, ch,x,y;

  if (iW>W) { sx= (iW-W)/2; cw= W; dx= 0; }
       else { sx= 0; cw= iW; dx= (W-iW)/2; }

  if (iH>H) { sy= (iH-H)/2; ch= H; dy= 0; }
       else { sy= 0; ch= iH; dy= 0; }

  for(y=0;y<ch;y++)
    for(x=0;x<cw;x++)
      D[ (x+dx)+(y+dy)*W ]= iD[ (x+sx)+ (y+sy)*iW ];
}

static void render_text
  (int fg, uint32_t *D,int W,int H,int tW,int tH, uint8_t *tD)
{
  int sx, sy, dx, dy, cw, ch,x,y;

  sx= 0; sy=0; dy= 0;

  if (tW>W) { dx= 0; cw= W; }
       else { dx= (W-tW)/2; cw= tW; }

  if (tH>H) ch= H; else ch= tH; 

  fg &= ~0xff000000;

  for(y=0;y<ch;y++)
    for(x=0;x<cw;x++)
      D[ (x+dx)+(y+dy)*W ]= (tD[ (x+sx)+ (y+sy)*tW ]<<24) | fg;
}

static char *extract_name(char *str)
{
  char *x;
  while((x=strchr(str,'/'))) str= x+1;
  str= strdup(str);
  x= strrchr(str,'.');
  if (x) *x= 0;
  return str;
}

static deskicon_t *make_icon_from_dir(deskfile_t *fil,cfg_desk_t *cfg)
{
  char *cmd;
  cmd= mprintf("%s %%f", cfg->G_file_manager ? cfg->file_manager : "xfe");
  return create_icon(cfg, cfg->G_directory_icon ? cfg->directory_icon : NULL,
                     ":file-dir.png", cmd, fil->name, fil->mtime);
}

static deskicon_t* make_icon_from_shortcut
   (deskfile_t *fil, cfg_desk_t *cfg)
{
  FILE *f;
  char *cmd, *ico;

  f= fopen(fil->name,"rb");
  if (!f) return NULL;

  cmd= ico= NULL;
  while(1)
  {
    char *key, *val, g[2000];
    int i;

    if (!fgets(g, sizeof(g),f)) break;
    g[sizeof(g)-1]= 0;
    for(i=0;g[i];i++) if (g[i]=='\r' || g[i]=='\n') { g[i]= 0; break; }
    key= g;
    val= strchr(g,'=');
    if (!val) continue;
    val[0]= 0;
    val++;
    key= trim_spaces(key);
    val= trim_spaces(val);
    if (!strcasecmp(key,"Exec")) cmd= strdup(val);
    if (!strcasecmp(key,"Icon")) ico= strdup(val);
  }
  fclose(f);

  if (ico) ico= find_icon_by_name(cfg,ico);

  return create_icon(cfg, ico, ":shortcut.png", cmd, fil->name, fil->mtime);
}

static int isspace(int chr)
{
  return ((chr>0) && (chr<=' ')) || (chr==127);
}

static char *trim_spaces(char *str)
{
  int L;
  while(isspace(*str)) str++;
  L= strlen(str);
  while(L>0)
    if (!isspace(str[L-1])) break;
    else str[L-1]= 0;
  return str;
}

static char *find_icon_by_name(cfg_desk_t *cfg, char *name)
{
  int i;
  char buf[2000];
  int n_icopa;
  char **icopa;

  if (name[0]=='/') return name;

  if (cfg->N_icon_path)
  {
   n_icopa= cfg->N_icon_path;
   icopa= cfg->icon_path;
  }
  else
  {
    static char *stdpath[]= {
     "/usr/share/icons/hicolor/32x32/actions",
     "/usr/share/icons/hicolor/32x32/animations",
     "/usr/share/icons/hicolor/32x32/apps",
     "/usr/share/icons/hicolor/32x32/categories",
     "/usr/share/icons/hicolor/32x32/devices",
     "/usr/share/icons/hicolor/32x32/emblems",
     "/usr/share/icons/hicolor/32x32/emotes",
     "/usr/share/icons/hicolor/32x32/filesystems",
     "/usr/share/icons/hicolor/32x32/intl",
     "/usr/share/icons/hicolor/32x32/mimetypes",
     "/usr/share/icons/hicolor/32x32/places",
     "/usr/share/icons/hicolor/32x32/status",
     "/usr/share/icons/hicolor/32x32/stock",
     "/usr/share/icons/hicolor/48x48/actions",
     "/usr/share/icons/hicolor/48x48/animations",
     "/usr/share/icons/hicolor/48x48/apps",
     "/usr/share/icons/hicolor/48x48/categories",
     "/usr/share/icons/hicolor/48x48/devices",
     "/usr/share/icons/hicolor/48x48/emblems",
     "/usr/share/icons/hicolor/48x48/emotes",
     "/usr/share/icons/hicolor/48x48/filesystems",
     "/usr/share/icons/hicolor/48x48/intl",
     "/usr/share/icons/hicolor/48x48/mimetypes",
     "/usr/share/icons/hicolor/48x48/places",
     "/usr/share/icons/hicolor/48x48/status",
     "/usr/share/icons/hicolor/48x48/stock" };
    icopa= stdpath;
    n_icopa= sizeof(stdpath)/sizeof(stdpath[0]);
  }

  for(i=0;i<n_icopa;i++)
  {
    snprintf(buf,sizeof(buf),"%s/%s.png", icopa[i], name);
    buf[sizeof(buf)-1]= 0;
    if (access(buf,R_OK)==0) return strdup(buf);
  }
  return NULL;
}

#line 377 "/home/dodo/tmp/mudesk-1.0/desk/mklist.u"

static int has_suffix(char *str,char *sfx)
{
  char *k;
  k= strrchr(str,'.');
  if (k) return !strcasecmp(k, sfx);
  return 0;
}

static int int_setting(int given, int setting, int defval, int min, int max)
{
  int V;
  V= given ? setting : defval;
  if (V<min || V>max) V= defval;
  return V;
}

static void find_file_type(char *fn, int nft, cfg_file_type_t *ft, 
                           char **Ricon, char **Rcmd)
{
  int i,j;
  for(i=0;i<nft;i++)
   for(j=0;j<ft[i].N_extensions;j++)
     if (has_suffix(fn, ft[i].extensions[j]))
     {
       if (Ricon) *Ricon= ft[i].icon;
       if (Rcmd) *Rcmd= ft[i].command;
       return ;
     }
  if (ft==builtin_ftype)
  {
     if (Ricon) *Ricon= ":file-generic.png";
     if (Rcmd)  *Rcmd= "xdg-open %f";
  }
}

