/*
Copyright (c) 2016 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#line 337 "/home/dodo/tmp/mudesk-1.0/desk/window.u"
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdint.h>
#include <stdio.h>
#include <syslog.h>
#include <X11/Xlib.h>
#include <X11/extensions/XShm.h>
#include "x11iface.h"
#include "config.h"
#include "bfont.h"
#include "txtren.h"
#include "menu.h"
#include "graymap.h"
#include "desk.h"
#include "mklist.h"
#include "window.h"
#include "pngio.h"
#include "resize_image.h"
#include "rect.h"
#include "main.h"

#line 35 "/home/dodo/tmp/mudesk-1.0/desk/window.u"

static Window win;
static int win_w, win_h;
static ximage_t *win_img;
static uint32_t *backgnd, back_w, back_h, back_x,back_y;
static int double_click_threshold= 250;
static int drag_threshold= 10;
static int clicked_icon=-1, click_x, click_y, 
           dragging, drag_delta_x,drag_delta_y,
           drag_last_x, drag_last_y;

#line 40 "window.c"

static void win_button_press(void *dummy, XEvent *event);
static void win_button_release(void *dummy, XEvent *event);
static void win_pointer_move(void *dummy, XEvent *event);
static void win_expose(void *dummy, XEvent *event);
static void load_background_image(cfg_desk_t *cfg);
 static void render_backgnd(int x,int y,int w,int h);
static void draw_icon(deskicon_t *ico,int x,int y,int w,int h);
static int clip_rect(int *Rx1,int *Ry1,int *Rw1,int *Rh1,
                     int x2, int y2, int w2, int h2);
static void update_window(int x,int y,int w,int h);
static int clip_for_rect(int *start, int *amount, int max);
static void drag_draw(int x,int y);
static void render_all();
static void window_load_config(cfg_desk_t *cfg);
#line 47 "/home/dodo/tmp/mudesk-1.0/desk/window.u"

static void win_button_press(void *dummy, XEvent *event)
{
  int x,y;
  x= event->xbutton.x;
  y= event->xbutton.y;
  //printf("button press event: %d %d (button %d)\n", 
        //event->xbutton.x, event->xbutton.y, event->xbutton.button);
  if (event->xbutton.button==3)
  {
    // printf("button 3\n");
    return ;
  }
  int i;
  clicked_icon= -1;
  for(i=0;i<n_icons;i++)
    if (x>=icons[i]->x && x<(icons[i]->x+icons[i]->w) &&
        y>=icons[i]->y && y<(icons[i]->y+icons[i]->h))
    {
      clicked_icon= i;
      break;
    }
  if (clicked_icon!=-1)
  {
    click_x= x;
    click_y= y;
    drag_delta_x= icons[i]->x - x;
    drag_delta_y= icons[i]->y - y;
  }
  dragging= 0;
}

static void win_button_release(void *dummy, XEvent *event)
{
  static Time last_press;
  Time delta;
  if (event->xbutton.button==3) 
  {
    int x,y,i;
    x= event->xbutton.x;
    y= event->xbutton.y;
    for(i=0;i<n_icons;i++)
      if (x>=icons[i]->x && x<(icons[i]->x+icons[i]->w) &&
          y>=icons[i]->y && y<(icons[i]->y+icons[i]->h))
      {
        show_item_menu(win, x,y, icons[i]->fn);
        return ;
      }
    show_window_menu(win,x,y);
    return ;
  }
  if (event->xbutton.button!=1) return ;
  if (dragging)
  {
    int ox,oy,w,h;
    int i;
    deskicon_t *CI;
    dragging= 0;
    CI= icons[clicked_icon];
    drag_draw(drag_last_x, drag_last_y);
    ox= CI->x;
    oy= CI->y;
    CI->x= event->xbutton.x + drag_delta_x;
    CI->y= event->xbutton.y + drag_delta_y;
    w= CI->w;
    h= CI->h;

    render_backgnd(ox,oy,w,h);
    render_backgnd(CI->x,CI->y,w,h);

    for(i=0;i<n_icons;i++)
      draw_icon(icons[i],ox,oy,w,h);

    for(i=0;i<n_icons;i++)
      draw_icon(icons[i],CI->x,CI->y,w,h);

    update_window(ox,oy,w,h);
    update_window(CI->x,CI->y,w,h);

    if (CI->encoded) { free(CI->encoded); CI->encoded= NULL; }
    write_desk();
    return ;
  }
  delta= event->xbutton.time-last_press;
  last_press= event->xbutton.time;
  if (delta<=double_click_threshold && clicked_icon!=-1)
  {
    execute_program(icons[clicked_icon]->cmd, icons[clicked_icon]->fn);
    clicked_icon= -1;
    return ;
  }
}

static void win_pointer_move(void *dummy, XEvent *event)
{
  int x,y,dist;
#if 0
  printf("pointer motion event: %d %d (button %d)\n", 
           event->xbutton.x, event->xbutton.y, event->xbutton.button);
#endif
  x= event->xmotion.x;
  y= event->xmotion.y;
  if (dragging)
  {
    drag_draw(drag_last_x,drag_last_y);
    drag_draw(x,y);
    drag_last_x= x;
    drag_last_y= y;
    // printf("drag %d %d\n",x,y);
    return ;
  }
  if (clicked_icon==-1) return ;
  dist= (x-click_x)*(x-click_x) + (y-click_y)*(y-click_y);
  if (dist>=drag_threshold*drag_threshold)
  {
    dragging= 1;
    drag_last_x= x;
    drag_last_y= y;
    drag_draw(x,y);
   // printf("drag start: %d %d\n",x,y);
  }
}


static void win_expose(void *dummy, XEvent *event)
{
  int x,y,w,h;
  x= event->xexpose.x;
  y= event->xexpose.y;
  w= event->xexpose.width;
  h= event->xexpose.height;
  put_image(win, x,y, win_img, x, y, w, h);
}

static void load_background_image(cfg_desk_t *cfg)
{
  char *e;
  unsigned int bw,bh,*bd;
  uint32_t rw, rh, *rd;

  if (backgnd) { free(backgnd); backgnd= NULL; }
  if (!cfg->G_background_image)
    return ;
  e= read_png_image(cfg->background_image, &bw,&bh,&bd);
  if (e) return ;
  if (bw<=win_w && bh<=win_h)
  {
   backgnd= bd;
    back_w= bw;
    back_h= bh;
    goto done;
  }
  if (resize_image_argb(bd, bw, bh, win_w, win_h,
                        &rd, &rw, &rh))
  {
     int x,y;
     back_w= bw > win_w ? win_w: bw;
     back_h= bh > win_h ? win_h: bh;
     backgnd= calloc(sizeof(backgnd[0]), back_w * back_h);
     for(y=0;y<back_h;y++)
       for(x=0;x<back_w;x++)
          backgnd[x+y*back_w]= bd[x+y*bw];
     free(bd);
     goto done;
  }
  free(bd);
  back_w= rw;
  back_h= rh;
  backgnd= rd;

done:
  back_x= (win_w-back_w)/2;
  back_y= (win_h-back_h)/2;
  return ;
}


#line 225 "/home/dodo/tmp/mudesk-1.0/desk/window.u"

// x,y,w,h are in window coordinates.
static void render_backgnd(int x,int y,int w,int h)
{
  rect_t req, img, paint[4];
  int i,n_paint;

  if (!backgnd)
  {
     // FIXME: the fill color should be configurable
    fill_rect(win_img->data, win_img->width, win_img->height, x,y,w,h, 0x00);
    return ;
  }

  rect(&img, back_x, back_y, back_x+back_w-1, back_y+back_h-1);
  rect(&req, x,y, x+w-1,y+h-1);

  n_paint= subtract_rectangle(img, &req, paint); 
  for(i=0;i<n_paint;i++)
    fill_rect(win_img->data, win_img->width, win_img->height, 
              paint[i].left, paint[i].top,
              paint[i].right-paint[i].left+1,
              paint[i].bot-paint[i].top+1, 0x00); // FIXME: fill color


  copy_area(win_img->data, win_img->width, win_img->height,
               req.left, req.top,
            backgnd, back_w, back_h,
               req.left-back_x, req.top-back_y,
            req.right-req.left+1, req.bot-req.top+1);
}

static void draw_icon(deskicon_t *ico,int x,int y,int w,int h)
{
  int cx,cy,cw,ch;
  cx= ico->x; cy= ico->y;
  cw= ico->w; ch= ico->h;
  if (clip_rect(&cx,&cy,&cw,&ch, x,y,w,h)) return ;
  draw_alpha_image(win_img->data, win_img->width, win_img->height,
                                  cx,cy,
                   ico->ico, ico->w, ico->h,
                                  cx-ico->x, cy-ico->y,
                                  cw,ch);
}

static int clip_rect(int *Rx1,int *Ry1,int *Rw1,int *Rh1,
                     int x2, int y2, int w2, int h2)
{
  int x1,y1,w1,h1;
  int r1,r2,b1,b2;
  x1= *Rx1; y1= *Ry1; w1= *Rw1; h1= *Rh1;

  r1= x1 + w1;
  r2= x2 + w2;
  b1= y1 + h1;
  b2= y2 + h2;

  if (x1>=r2) return 1;
  if (r1<=x2) return 1;
  if (y1>=b2) return 1;
  if (b1<=y2) return 1;

  if (r1>r2) r1= r2;
  if (b1>b2) b1= b2;

  if (x1<x2) x1= x2;
  if (y1<y2) y1= y2;

  w1= r1 - x1;
  h1= b1 - y1;

  if (w1<=0 || h1<=0) return 1;

  *Rx1= x1; *Ry1= y1; *Rw1= w1; *Rh1= h1;
  return 0;
}

static void update_window(int x,int y,int w,int h)
{
  if (clip_for_rect(&x,&w, win_w) || 
      clip_for_rect(&y,&h, win_h)) return ;
  XClearArea(x_display, win, x,y,w,h, True);
}
#line 309 "/home/dodo/tmp/mudesk-1.0/desk/window.u"

static int clip_for_rect(int *start, int *amount, int max)
{
  int end;
  if (*amount<=0) return 1;
  end= *start + *amount;
  if (end>max) end= max;
  if (*start<0) *start= 0;
  *amount= end - *start;
  return *amount<=0;
}

static void drag_draw(int x,int y)
{
  int sx,sy;
  int w,h;
  sx= drag_delta_x + x;
  sy= drag_delta_y + y;

  w= icons[clicked_icon]->w;
  h= icons[clicked_icon]->h;

  xor_rect(win_img->data,win_img->width, win_img->height,sx, sy,w,h, 0xFFffFF);
  update_window(sx,sy,w,h);
}


#line 411 "/home/dodo/tmp/mudesk-1.0/desk/window.u"

static void render_all()
{
  int i;
  render_backgnd(0,0,win_w,win_h);
  for(i=0;i<n_icons;i++)
    draw_icon(icons[i],0,0,win_w,win_h);
  update_window(0,0,win_w,win_h);
}

static void window_load_config(cfg_desk_t *cfg)
{
  if (cfg->G_double_click_threshold)
    double_click_threshold= cfg->double_click_threshold;
}

#line 10 "/home/dodo/tmp/mudesk-1.0/desk/window.u"

void init_window(cfg_desk_t* cfg,int width,int height, int dbgwin)
{
  int events;

  if (width<0) width= x_root_width;
  if (height<0) height= x_root_height;

  events= ExposureMask | ButtonPressMask | ButtonReleaseMask 
                                         | Button1MotionMask;

  win= create_window(x_root_window,0,0,width,height,events, 0, !dbgwin);

  win_w= width;
  win_h= height;

  register_event_handler(win, MotionNotify, NULL, win_pointer_move, NULL);
  register_event_handler(win, ButtonRelease, NULL, win_button_release, NULL);
  register_event_handler(win, ButtonPress, NULL, win_button_press, NULL);
  register_event_handler(win, Expose, NULL, win_expose, NULL);

  win_img= create_image(win_w, win_h);
}

#line 376 "/home/dodo/tmp/mudesk-1.0/desk/window.u"

void replace_icon(char *fn, deskicon_t *ico)
{
  deskicon_t *old;
  int i;

  for(i=0;i<n_icons;i++)
   if (!strcmp(icons[i]->fn,fn)) break;
  if (i==n_icons)
  {
    deskicon_free(ico);
    return ;
  }
  old= icons[i];
  icons[i]= ico;
  ico->x= old->x;
  ico->y= old->y;

  render_backgnd(old->x,old->y,old->w,old->h);
  render_backgnd(ico->x,ico->y,ico->w,ico->h);
  for(i=0;i<n_icons;i++)
    if (icons[i]!=old)
    {
      draw_icon(icons[i],old->x,old->y,old->w,old->h);
      draw_icon(icons[i],ico->x,ico->y,ico->w,ico->h);
    }
  draw_icon(ico,ico->x,ico->y,ico->w,ico->h);

  update_window(old->x,old->y,old->w,old->h);
  update_window(ico->x,ico->y,ico->w,ico->h);

  deskicon_free(old);
}

#line 428 "/home/dodo/tmp/mudesk-1.0/desk/window.u"

void window_restart(cfg_desk_t *cfg)
{
  window_load_config(cfg);
  load_background_image(cfg);
  render_all();
}
