/*
Copyright (c) 2016 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#ifndef x_basics_h_included_lskdsoweitjkjsd
#define x_basics_h_included_lskdsoweitjkjsd


/***********************************************************************
 *                                                                     *
 *                               BASICS                                *
 *                                                                     *
 ***********************************************************************/

/****
    - opens the display
    - finds a 24-bit visual
    - gets various properties
      -- root window size
      -- pixel format
****/

extern Display *x_display;
extern Window   x_root_window;
extern int      x_screen_number;
extern uint32_t x_root_width, x_root_height;
extern uint32_t x_red_mask, x_red_shift,
                x_green_mask, x_green_shift,
                x_blue_mask, x_blue_shift;
extern int      x_depth;
extern Visual  *x_visual;

int init_display(char *name);

Window create_window
              (Window parent, int x,int y, int w, int h,
               unsigned int events, int border, int override);


/***********************************************************************
 *                                                                     *
 *                            EVENT SYSTEM                             *
 *                                                                     *
 ***********************************************************************/

void register_event_handler(Window window, unsigned int event,
                         void *data, void *func, void**old);
void unregister_event_handler(Window window, unsigned int event, void *old);
void unregister_all_event_handlers(Window window);
void handle_registered_event(XEvent *event);

void call_event_handler(void *H,XEvent *event);
void free_event_handler(void *H);

int grab_key(char *keyspec, int *Rmod, int *Rkey,
             void *press_func, void *release_func, void *arg);

int grab_keyboard(void *press_func, void *release_func, void *arg);
void ungrab_keyboard();

void grab_extra_modifiers(char *exmod);


/***********************************************************************
 *                                                                     *
 *                           IMAGE HANDLING                            *
 *                                                                     *
 ***********************************************************************/

typedef struct 
{
  uint32_t *data;
  unsigned int width,height, bpl;
  int shared;
  XShmSegmentInfo shminfo;
  XImage *x11image;
} ximage_t;


ximage_t *create_image(uint32_t width, uint32_t height);
void      destroy_image(ximage_t *image);
void      put_image(Window window, int dest_x, int dest_y,
                    ximage_t *image, int src_x, int src_y,
                    int width, int height);
void      handle_image_chores();

#endif
